const TrekRouter = require('@beanjs/trek-router')
const flows = require('./flows')

const {
  kTeamitySockets,
  kTeamityReply,
  kTeamitySerializer,
  kSocketTeamity,

  kRouteScope,
  kRouteSocket,
  kRouteParent,
  kRouteReply,
  kRouteAttr,
  kRouteAjvParams,
  kRouteAjvBody,
  kRouteAjvReply,

  kReplyRoute,
  kScopeBody,
  kScopeParams
} = require('./symbols')
const { RouteExistsError } = require('./errors')
const { Socket } = require('./socket')
const { Route } = require('./route')
const { Scope } = require('./scope')
const { initReply } = require('./reply')
const {
  throwError,
  onBeforeHandlerFlow,
  onAfterHandlerFlow,
  onBeforeConnectionFlow,
  onAfterDisconnectionFlow
} = require('./hooks')

const { buildAjvErrorsMsg } = require('./utils')

const router = new TrekRouter({
  SLASH: '.'
})
const method = 'TEAMITY'

function addSocket () {
  const teamity = this[kSocketTeamity]
  const sockets = teamity[kTeamitySockets]
  sockets[this.$id] = this

  flows.series(this, [onBeforeConnectionFlow], err => {
    if (err) {
      throwError(teamity, err)
      this.$raw.close()
    }
  })
}

function delSocket () {
  const teamity = this[kSocketTeamity]
  const sockets = teamity[kTeamitySockets]
  delete sockets[this.$id]
  this.$raw.close()

  flows.series(this, [onAfterDisconnectionFlow], err => {
    if (err) {
      throwError(teamity, err)
    }
  })
}

function recvSocket (url, body, attr) {
  const findResult = router.find(method, url)
  if (!(findResult[0] instanceof Route)) {
    return
  }

  const parentRoute = findResult[0]
  const { $teamity } = parentRoute
  const params = {}
  findResult[1].forEach(pm => {
    params[pm.name] = pm.value
  })

  const route = Object.create(parentRoute)
  const rep = Object.create($teamity[kTeamityReply])
  rep[kReplyRoute] = route
  initReply.call(rep, this)
  const scope = new Scope(this, url, params, body || {})

  route[kRouteParent] = parentRoute
  route[kRouteScope] = scope

  onBeforeHandlerFlow.call(route, body, attr, err => {
    if (err) {
      return throwError($teamity, err)
    }

    route[kRouteSocket] = this
    route[kRouteReply] = rep
    route[kRouteAttr] = attr

    onHandlerFlow.call(route, err => {
      if (err) {
        throwError($teamity, err)
      }
    })
  })
}

function addRouteFlow (next) {
  const { url, method, $teamity, schema } = this
  const { $log } = $teamity

  const findResult = router.find(method, url)
  if (findResult[0]) {
    return next(new RouteExistsError(url))
  }

  if (schema && schema.params) {
    this[kRouteAjvParams] = $teamity.$ajv.compile(schema.params)
  }

  if (schema && schema.body) {
    this[kRouteAjvBody] = $teamity.$ajv.compile(schema.body)
  }

  if (schema && schema.reply) {
    this[kRouteAjvReply] = $teamity.$ajv.compile(schema.reply)
  }

  $log.info(`teamity route: ${url}`)
  router.add(method, url, this)
  next()
}

async function onHandlerFlow (next) {
  const { handler, $teamity } = this
  const scope = this[kRouteScope]
  const rep = this[kRouteReply]

  try {
    const ajvBodyCompiler = this[kRouteAjvBody]
    const ajvParamsCompiler = this[kRouteAjvParams]

    if (ajvBodyCompiler && !ajvBodyCompiler(scope[kScopeBody])) {
      throw new Error(buildAjvErrorsMsg(ajvBodyCompiler.errors))
    }

    if (ajvParamsCompiler && !ajvParamsCompiler(scope[kScopeParams])) {
      throw new Error(buildAjvErrorsMsg(ajvParamsCompiler.errors))
    }

    handler.call(this, scope, rep)
  } catch (e) {
    throwError($teamity, e)
  }
  next()
}

function onWsComing (ws, req) {
  const { query, session = null } = req
  const skt = new Socket(ws, query, session)
  skt[kSocketTeamity] = this

  skt.on('close', delSocket)
  skt.on('error', delSocket)
  skt.on('decode', recvSocket)
  addSocket.call(skt)
}

async function onWsSending ({ url, reply, route, sockets }, done) {
  try {
    const serializer = this[kTeamitySerializer]
    if (typeof serializer !== 'function') {
      this.$log.error('No setting serializer!!')
      return done()
    }

    function _wsSending (attr) {
      serializer.call(this, url, reply, attr, (err, payload) => {
        if (err) {
          return throwError(this, err)
        }

        for (const skt of sockets) {
          skt.$raw.send(payload)
        }
      })
    }

    if (route) {
      const ajvReplyCompiler = route[kRouteAjvReply]
      if (ajvReplyCompiler && !ajvReplyCompiler(reply)) {
        throw new Error(buildAjvErrorsMsg(ajvReplyCompiler.errors))
      }

      onAfterHandlerFlow.call(route, reply, route[kRouteAttr], err => {
        if (err) {
          return throwError(this, err)
        }

        _wsSending(route[kRouteAttr])
      })
    } else {
      _wsSending()
    }
  } catch (e) {
    throwError(this, e)
  }
  done()
}

module.exports = {
  addRouteFlow,
  onWsComing,
  onWsSending
}
